_CONNECTION_STRING = "InstrumentationKey=4128b5ed-f5ec-4612-ac32-b8cff0f2de93;IngestionEndpoint=https://northeurope-2.in.applicationinsights.azure.com/;LiveEndpoint=https://northeurope.livediagnostics.monitor.azure.com/;ApplicationId=f448014b-2abf-4681-8e48-583e33f7d173"

_APP_INSIGHTS_EVENT_MARKER_ATTRIBUTE = "APPLICATION_INSIGHTS_EVENT_MARKER_ATTRIBUTE"
_OTEL_RESOURCE_ATTRIBUTES = "OTEL_RESOURCE_ATTRIBUTES"
_SDK_VERSION = "SdkVersion"
_PROJECT_KEY = "ProjectKey"

_TELEMETRY_CONFIG_FILE = ".telemetry.json"

_CODE_FILEPATH = "code.file.path"
_CODE_FUNCTION = "code.function.name"
_CODE_LINENO = "code.line.number"

_CLOUD_ORG_ID = "CloudOrganizationId"
_CLOUD_TENANT_ID = "CloudTenantId"
_CLOUD_URL = "CloudUrl"
_APP_NAME = "ApplicationName"

_UNKNOWN = ""
